#pragma once

#include <cctype>
#include <string>
#include <memory>
#include <fstream>
#include <algorithm>

namespace util
{
    // this class exists so that from_string is able to use stringstreams to 
    // convert values. the problem is the >> operator of stringstreams overwrites
    // the contents of the target - this is fine for other basic types such as 
    // int, but with strings we want it to concatenate
    template<typename T> class ValueProxy
    {
    public:
        friend std::stringstream& operator >> (std::stringstream& out, 
            ValueProxy& valueProxy)
        {
            valueProxy.SetValue(out);
            return out;
        }

        void SetValue(std::stringstream& out)
        {
            m_stringValue = out.str();
            out >> m_value;
        }

        template<typename U, typename = typename std::enable_if<
            !std::is_same<std::string, T>::value, T>::type> 
        void AssignTo(U& value)
        {
            value = m_value;
        }

        void AssignTo(std::string& value)
        {
            value = m_stringValue;
        }

    private:
        T m_value;
        std::string m_stringValue;
    };

    template<typename T> inline std::string toStr(const T& tmp)
    { 
        std::string convertedValue;
        std::stringstream s;
        s << tmp;

        ValueProxy<std::string> value;
        s >> value;
        value.AssignTo(convertedValue);

        return convertedValue; 
    }

    template<typename T> std::string toStr(T value, unsigned int precision)
    {
        std::ostringstream o;

        if(precision == 0)
        {
            o << value;
        }
        else
        {
            o << std::setfill('0') << std::fixed << std::setprecision(precision) 
                << value;
        }

        return o.str();
    }

    template<typename T> inline T fromStr(const std::string& str) 
    {
        T convertedValue; 
        std::stringstream s;
        s << str; 

        ValueProxy<T> value;
        s >> value;
        value.AssignTo(convertedValue);

        return convertedValue; 
    }

    static inline std::wstring toWstr(const std::string& s)
    {
        if(true == s.empty())
        {
            return std::wstring();
        }
        // overestimate number of code points
        std::wstring ws(s.size() + 1, L' '); 
        size_t convertedCharCount;
        mbstowcs_s(&convertedCharCount, &ws[0], ws.size(), s.c_str(), s.size());
        // shrink to fit
        ws.resize(convertedCharCount); 
        return ws;
    }
    // trim from start
    static inline std::string& lTrim(std::string& s) 
    {
        s.erase(s.begin(), std::find_if(
            s.begin(), s.end(), std::not1(
                std::ptr_fun<int, int>(std::isspace))));

        return s;
    }
    // trim from end
    static inline std::string& rTrim(std::string& s) 
    {
        s.erase(std::find_if(s.rbegin(), s.rend(), std::not1(
            std::ptr_fun<int, int>(std::isspace))).base(), s.end());

        return s;
    }
    // trim from both ends
    static inline std::string& trim(std::string& s) 
    {
        return lTrim(rTrim(s));
    }

    static inline std::string stripChars(const std::string& inputStr, 
        const std::string& charsToRemove)
    {
        std::string resultStr = inputStr;
        for(unsigned int i = 0; i < charsToRemove.size(); ++i)
        {
            resultStr.erase(
                std::remove(
                    resultStr.begin(), resultStr.end(), charsToRemove[i]),
                resultStr.end());
        }

        return resultStr;
    }
    // capitalise the first letter of the supplied string
    static inline std::string capitalise(const std::string& s)
    {
        if(s.size() == 0)
        {
            return s;
        }

        std::string r = s;
        r[0] = toupper(r[0]);
        return r;
    }

    static inline std::string toLower(const std::string& s)
    {
        std::string result = s;
        std::transform(result.begin(), result.end(), result.begin(), std::tolower);
        return result;
    }
    // returns -1 if the delimiter was not found the specified number of times
    static inline int findNth(const std::string& inputStr, char delim, 
        int count, unsigned int startOffset = 0)
    {
        int offset = -1;
        int foundCount = 0;

        for(unsigned int i = startOffset; i < inputStr.length(); ++i)
        {
            char currChar = inputStr[i];

            if(currChar == delim)
            {
                ++foundCount;
            }

            if(foundCount == count)
            {
                offset = i;
                break;
            }
        }

        return offset;
    }

    static inline bool startsWith(const std::string& str, 
        const std::string& preFix)
    {
        int strLen = preFix.size();
        return str.substr(0, strLen) == preFix;
    }

    static inline bool endsWith(const std::string& str, 
        const std::string& postFix)
    {
        int strLen = str.size();
        int postFixLen = postFix.size();

        if(strLen < postFixLen)
        {
            return false;
        }

        return str.substr(strLen - postFixLen) == postFix;
    }
    // strip basic authentication data from supplied url
    static inline std::string stripCredentials(const std::string& url)
    {
        int credentialsStart = url.find_first_of("//") + 2;
        int credentialsEnd = url.find_first_of('@') + 1;

        if(credentialsStart == std::string::npos 
        || credentialsEnd == std::string::npos)
        {
            return url;
        }

        return url.substr(0, credentialsStart) + url.substr(credentialsEnd);
    }
}
